from django.db import models


class Person(models.Model):
    name = models.TextField()

    def __str__(self):
        return f"{self.name} ({self.pk})"


class PersonContent(models.Model):
    person = models.ForeignKey(Person, on_delete=models.CASCADE)
    language = models.TextField()
    text = models.TextField()

    def __str__(self):
        return self.text

    def get_absolute_url(self):
        return "/"



from cms import constants as cms_constants
from cms.apphook_pool import apphook_pool
from cms.cms_menus import CMSMenu as OriginalCMSMenu, get_visible_nodes
from cms.models import Page
from cms.toolbar.utils import get_object_preview_url, get_toolbar_from_request
from cms.utils.page import get_page_queryset
from django.apps import apps
from menus.base import Menu, NavigationNode
from menus.menu_pool import menu_pool

from . import conf, constants


class CMSVersionedNavigationNode(NavigationNode):
    def is_selected(self, request):
        try:
            page_id = request.current_page.pk
        except AttributeError:
            return False
        return page_id == self.id


def _get_attrs_for_node(renderer, page_content):
    page = page_content.page
    language = renderer.request_language
    attr = {
        "is_page": True,
        "soft_root": page_content.soft_root,
        "auth_required": page.login_required,
        "reverse_id": page.reverse_id,
    }
    limit_visibility_in_menu = page_content.limit_visibility_in_menu

    if limit_visibility_in_menu is cms_constants.VISIBILITY_ALL:
        attr["visible_for_authenticated"] = True
        attr["visible_for_anonymous"] = True
    else:
        attr["visible_for_authenticated"] = (
            limit_visibility_in_menu == cms_constants.VISIBILITY_USERS
        )
        attr["visible_for_anonymous"] = (
            limit_visibility_in_menu == cms_constants.VISIBILITY_ANONYMOUS
        )

    attr["is_home"] = page.is_home
    extenders = []

    if page.navigation_extenders:
        if page.navigation_extenders in renderer.menus:
            extenders.append(page.navigation_extenders)
        elif f"{page.navigation_extenders}:{page.pk}" in renderer.menus:
            extenders.append(f"{page.navigation_extenders}:{page.pk}")

    if page.application_urls:
        app = apphook_pool.get_apphook(page.application_urls)

        if app:
            extenders += app.get_menus(page, language)

    exts = []

    for ext in extenders:
        if hasattr(ext, "get_instances"):
            exts.append(f"{ext.__name__}:{page.pk}")
        elif hasattr(ext, "__name__"):
            exts.append(ext.__name__)
        else:
            exts.append(ext)

    if exts:
        attr["navigation_extenders"] = exts

    attr["redirect_url"] = page_content.redirect

    return attr


class CMSMenu(Menu):
    def get_nodes(self, request):
        site = self.renderer.site
        language = self.renderer.request_language
        pages_qs = get_page_queryset(site).select_related("node")
        visible_pages_for_user = get_visible_nodes(request, pages_qs, site)

        if not visible_pages_for_user:
            return []

        cms_extension = apps.get_app_config("djangocms_versioning").cms_extension
        toolbar = get_toolbar_from_request(request)
        edit_or_preview = toolbar.edit_mode_active or toolbar.preview_mode_active
        menu_nodes = []
        node_id_to_page = {}
        homepage_content = None

        # Depending on the toolbar mode, we need to get the correct version.
        # On edit or preview mode: return DRAFT,
        # if DRAFT does not exist then return PUBLISHED.
        # On public mode: return PUBLISHED.
        if edit_or_preview:
            states = [constants.DRAFT, constants.PUBLISHED]
        else:
            states = [constants.PUBLISHED]

        versionable_item = cms_extension.versionables_by_grouper[Page]
        versioned_page_contents = (
            versionable_item.content_model._base_manager.filter(
                language=language, page__in=pages_qs, versions__state__in=states
            )
            .order_by("page__node__path", "versions__state")
            .select_related("page", "page__node")
            .prefetch_related("versions")
        )
        added_pages = []

        for page_content in versioned_page_contents:
            page = page_content.page

            if page not in visible_pages_for_user:
                # The page is restricted for the user.
                # Therefore, we avoid adding it to the menu.
                continue

            version = page_content.versions.all()[0]

            if (
                page.pk in added_pages
                and edit_or_preview
                and version.state == constants.PUBLISHED
            ):
                # Page content is already added. This is the case where you
                # have both draft and published and in edit/preview mode.
                # We give priority to draft which is already sorted by the query.
                # Therefore we ignore the published version.
                continue

            page_tree_node = page.node
            parent_id = node_id_to_page.get(page_tree_node.parent_id)

            if page_tree_node.parent_id and not parent_id:
                # If the parent page is not available,
                # we skip adding the menu node.
                continue

            # Construct the url based on the toolbar mode.
            if edit_or_preview:
                url = get_object_preview_url(page_content)
            else:
                url = page_content.get_absolute_url()

            # Create the new navigation node.
            new_node = CMSVersionedNavigationNode(
                id=page.pk,
                attr=_get_attrs_for_node(self.renderer, page_content),
                title=page_content.menu_title or page_content.title,
                url=url,
                visible=page_content.in_navigation,
            )

            if not homepage_content:
                # Set the home page content.
                homepage_content = page_content if page.is_home else None

            cut_homepage = homepage_content and not homepage_content.in_navigation

            if cut_homepage and parent_id == homepage_content.page.pk:
                # When the homepage is hidden from navigation,
                # we need to cut all its direct children from it.
                new_node.parent_id = None
            else:
                new_node.parent_id = parent_id

            node_id_to_page[page_tree_node.pk] = page.pk
            menu_nodes.append(new_node)
            added_pages.append(page.pk)
        return menu_nodes


if conf.ENABLE_MENU_REGISTRATION:
    # Remove the core djangoCMS CMSMenu and register the new CMSVersionedMenu.
    menu_pool.menus.pop(OriginalCMSMenu.__name__)
    menu_pool.register_menu(CMSMenu)



from cms.admin.utils import GrouperModelAdmin
from django.contrib import admin

from djangocms_versioning.admin import (
    ExtendedGrouperVersionAdminMixin,
    ExtendedVersionAdminMixin,
    StateIndicatorMixin,
)
from djangocms_versioning.test_utils.blogpost import models


class BlogContentAdmin(StateIndicatorMixin, ExtendedVersionAdminMixin, admin.ModelAdmin):
    list_display = ("__str__", "state_indicator")


class BlogPostAdmin(StateIndicatorMixin, ExtendedGrouperVersionAdminMixin, GrouperModelAdmin):
    content_model = models.BlogContent  # Non-standard naming
    grouper_field_name = "blogpost"
    list_display = ("__str__", "state_indicator")


admin.site.register(models.BlogPost, BlogPostAdmin)
admin.site.register(models.BlogContent, BlogContentAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-08-20 15:54
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("djangocms_versioning", "0007_auto_20180813_1407")]

    operations = [
        migrations.AlterField(
            model_name="version",
            name="created_by",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to=settings.AUTH_USER_MODEL,
                verbose_name="author",
            ),
        )
    ]



from django import template

register = template.Library()


@register.inclusion_tag("polls/poll_tag.html", takes_context=True)
def render_poll(context, instance):
    pollcontent = instance.pollcontent_set.first()
    return {"content": pollcontent}



from cms.extensions.models import BaseExtension
from cms.operations import (
    ADD_PLUGIN,
    ADD_PLUGINS_FROM_PLACEHOLDER,
    CHANGE_PLUGIN,
    CLEAR_PLACEHOLDER,
    CUT_PLUGIN,
    DELETE_PLUGIN,
    MOVE_PLUGIN,
    PASTE_PLACEHOLDER,
    PASTE_PLUGIN,
)
from django.utils import timezone

from .models import Version
from .versionables import _cms_extension


def _update_modified(instance):
    if isinstance(instance, BaseExtension):
        instance = instance.extended_object
    if instance and _cms_extension().is_content_model_versioned(instance.__class__):
        try:
            version = Version.objects.get_for_content(instance)
        except Version.DoesNotExist:
            return
        version.modified = timezone.now()
        version.save(update_fields=["modified"])


def update_modified_date(sender, **kwargs):
    if kwargs["created"]:
        return
    _update_modified(kwargs["instance"])


def update_modified_date_for_pagecontent(sender, **kwargs):
    instance = kwargs["obj"].get_content_obj()
    _update_modified(instance)


def update_modified_date_for_placeholder_source(sender, **kwargs):
    placeholders = []
    operation = kwargs["operation"]
    if operation in (ADD_PLUGIN, CHANGE_PLUGIN, CLEAR_PLACEHOLDER, DELETE_PLUGIN):
        placeholders = [kwargs["placeholder"]]
    elif operation in (ADD_PLUGINS_FROM_PLACEHOLDER, PASTE_PLACEHOLDER, PASTE_PLUGIN):
        placeholders = [kwargs["target_placeholder"]]
    elif operation in (CUT_PLUGIN,):
        placeholders = [kwargs["source_placeholder"]]
    elif operation in (MOVE_PLUGIN,):
        placeholders = [kwargs["source_placeholder"], kwargs["target_placeholder"]]

    for placeholder in placeholders:
        _update_modified(placeholder.source)



import collections

from cms.app_base import CMSAppConfig, CMSAppExtension
from cms.extensions.models import BaseExtension
from cms.models import PageContent, Placeholder
from cms.utils import get_language_from_request
from cms.utils.i18n import get_language_list, get_language_tuple
from cms.utils.plugins import copy_plugins_to_placeholder
from cms.utils.urlutils import admin_reverse
from django.conf import settings
from django.contrib.admin.utils import flatten_fieldsets
from django.core.exceptions import (
    ImproperlyConfigured,
    ObjectDoesNotExist,
    PermissionDenied,
)
from django.http import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
)
from django.utils.encoding import force_str
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from . import indicators, versionables
from .admin import VersioningAdminMixin
from .constants import INDICATOR_DESCRIPTIONS
from .datastructures import BaseVersionableItem, VersionableItem
from .exceptions import ConditionFailed
from .helpers import (
    get_latest_admin_viewable_content,
    inject_generic_relation_to_version,
    is_editable,
    register_versionadmin_proxy,
    replace_admin_for_models,
    replace_manager,
)
from .managers import AdminManagerMixin, PublishedContentManagerMixin
from .models import Version
from .plugin_rendering import CMSToolbarVersioningMixin


class VersioningCMSExtension(CMSAppExtension):
    def __init__(self):
        self.versionables = []
        self.add_to_context = {}
        self.add_to_field_extension = {}

    @cached_property
    def versionables_by_content(self):
        """Returns a dict of {content_model_cls: VersionableItem obj}"""
        return {
            versionable.content_model: versionable for versionable in self.versionables
        }

    def is_content_model_versioned(self, content_model):
        """Returns if the content model is registered for versioning.
        """
        return content_model in self.versionables_by_content

    @cached_property
    def versionables_by_grouper(self):
        """Returns a dict of {grouper_model_cls: VersionableItem obj}"""
        return {
            versionable.grouper_model: versionable
            for versionable in self.versionables
            # TODO: Comment on/document why this is here
            if versionable.concrete
        }

    def is_grouper_model_versioned(self, grouper_model):
        """Returns if the grouper model has been registered for versioning"""
        return grouper_model in self.versionables_by_grouper

    def handle_versioning_setting(self, cms_config):
        """Check the versioning setting has been correctly set
        and add it to the masterlist if all is ok
        """
        # First check that versioning is correctly defined
        if not isinstance(cms_config.versioning, collections.abc.Iterable):
            raise ImproperlyConfigured("versioning not defined as an iterable")
        for versionable in cms_config.versioning:
            if not isinstance(versionable, BaseVersionableItem):
                raise ImproperlyConfigured(
                    f"{versionable!r} is not a subclass of djangocms_versioning.datastructures.BaseVersionableItem"
                )
            # NOTE: Do not use the cached property here as this is
            # still changing and needs to be calculated on the fly
            registered_so_far = [v.content_model for v in self.versionables]
            if versionable.content_model in registered_so_far:
                raise ImproperlyConfigured(
                    f"{versionable.content_model!r} has already been registered"
                )
            # Checks passed. Add versionable to our master list
            self.versionables.append(versionable)

    def handle_versioning_add_to_confirmation_context_setting(self, cms_config):
        """
        Check versioning_add_to_confirmation_context has been correctly set
        and add it to the master dict if all is ok
        """
        add_to_context = cms_config.versioning_add_to_confirmation_context
        supported_keys = ["unpublish"]
        for key, value in add_to_context.items():
            if key not in supported_keys:
                raise ImproperlyConfigured(
                    f"{key!r} is not a supported dict key in the versioning_add_to_confirmation_context setting"
                )
            if key not in self.add_to_context:
                self.add_to_context[key] = collections.OrderedDict()
            self.add_to_context[key].update(value)

    def handle_admin_classes(self, cms_config):
        """Replaces admin model classes for all registered content types
        with an admin model class that inherits from `versionable.content_admin_mixin`.
        """
        replace_admin_for_models(
            [
                (versionable.content_model, versionable.content_admin_mixin)
                for versionable in cms_config.versioning
            ]
        )

    def handle_version_admin(self, cms_config):
        """
        Registers version admin for all registered content types
        with filtering by content type applied, so only versions for
        that specific content type are shown.
        """
        for versionable in cms_config.versioning:
            if versionable.concrete:
                register_versionadmin_proxy(versionable)

    def handle_content_model_generic_relation(self, cms_config):
        """Adds `versions` GenericRelation field to all provided
        content models.
        """
        for versionable in cms_config.versioning:
            inject_generic_relation_to_version(versionable.content_model)

    def handle_content_model_manager(self, cms_config):
        """Replaces default manager in provided content models with
        one inheriting from PublishedContentManagerMixin.
        """
        for versionable in cms_config.versioning:
            replace_manager(versionable.content_model, "objects", PublishedContentManagerMixin)
            replace_manager(versionable.content_model, "admin_manager", AdminManagerMixin,
                            _group_by_key=list(versionable.grouping_fields))

    def handle_admin_field_modifiers(self, cms_config):
        """Allows for the transformation of a given field in the ExtendedVersionAdminMixin
        """
        extended_admin_field_modifiers = getattr(cms_config, "extended_admin_field_modifiers", None)
        if not isinstance(extended_admin_field_modifiers, list):
            raise ImproperlyConfigured("extended_admin_field_modifiers must be list of dictionaries")
        for modifier in extended_admin_field_modifiers:
            for key in modifier.keys():
                self.add_to_field_extension[key] = modifier[key]

    def configure_app(self, cms_config):
        if hasattr(cms_config, "extended_admin_field_modifiers"):
            self.handle_admin_field_modifiers(cms_config)
        # Validation to ensure either the versioning or the
        # versioning_add_to_confirmation_context config has been defined
        has_extra_context = hasattr(
            cms_config, "versioning_add_to_confirmation_context"
        )
        has_models_to_register = hasattr(cms_config, "versioning")
        if not has_extra_context and not has_models_to_register:
            raise ImproperlyConfigured(
                "The versioning or versioning_add_to_confirmation_context setting must be defined"
            )
        # No exception raised so now configure based on those settings
        if has_extra_context:
            self.handle_versioning_add_to_confirmation_context_setting(cms_config)
        if has_models_to_register:
            self.handle_versioning_setting(cms_config)
            self.handle_admin_classes(cms_config)
            self.handle_version_admin(cms_config)
            self.handle_content_model_generic_relation(cms_config)
            self.handle_content_model_manager(cms_config)


def copy_page_content(original_content):
    """Copy the PageContent object and deepcopy its
    placeholders and plugins.
    """
    # Copy content object
    content_fields = {
        field.name: getattr(original_content, field.name)
        for field in PageContent._meta.fields
        # Don't copy the pk as we're creating a new obj.
        # The creation date should reflect the date it was copied on,
        # so don't copy that either.
        if field.name not in (PageContent._meta.pk.name, "creation_date")
    }

    # Use original manager to not create a new Version object here
    new_content = PageContent._original_manager.create(**content_fields)

    # Copy placeholders
    new_placeholders = []
    for placeholder in original_content.placeholders.all():
        placeholder_fields = {
            field.name: getattr(placeholder, field.name)
            for field in Placeholder._meta.fields
            # don't copy primary key because we're creating a new obj
            # and handle the source field later
            if field.name not in [Placeholder._meta.pk.name, "source"]
        }
        if placeholder.source:
            placeholder_fields["source"] = new_content
        new_placeholder = Placeholder.objects.create(**placeholder_fields)
        # Copy plugins
        placeholder.copy_plugins(new_placeholder)
        new_placeholders.append(new_placeholder)
    new_content.placeholders.add(*new_placeholders)

    # If pagecontent has an associated content or page extension, also copy this!
    for field in PageContent._meta.related_objects:
        if hasattr(original_content, field.name):
            extension = getattr(original_content, field.name)
            if isinstance(extension, BaseExtension):
                extension.copy(new_content, new_content.language)

    return new_content


def label_from_instance(obj, language):
    """
    Override the label for each grouper select option
    """
    title = obj.get_title(language) or _("No available title")
    path = obj.get_path(language)
    path = f"/{path}/" if path else _("Unpublished")
    return f"{title} ({path})"


def on_page_content_publish(version):
    """Url path and cache operations to do when a PageContent obj is published"""
    page = version.content.page
    language = version.content.language
    page._update_url_path(language)
    if page.is_home:
        page._remove_title_root_path()
    page._update_url_path_recursive(language)
    page.clear_cache(menu=True)

def on_page_content_unpublish(version):
    """Url path and cache operations to do when a PageContent obj is unpublished"""
    page = version.content.page
    language = version.content.language
    page._update_url_path_recursive(language)
    page.clear_cache(menu=True)


def on_page_content_draft_create(version):
    """Clear cache when a new PageContent draft is created."""
    page = version.content.page
    page.clear_cache(menu=True)


def on_page_content_archive(version):
    """Clear cache when a new PageContent version is archived."""
    page = version.content.page
    page.clear_cache(menu=True)


class VersioningCMSPageAdminMixin(VersioningAdminMixin):
    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)
        if obj:
            version = Version.objects.get_for_content(obj)
            if not version.check_modify.as_bool(request.user):
                form = self.get_form_class(request)
                if form.fieldsets:
                    fields = flatten_fieldsets(form.fieldsets)
                fields = list(fields)
                for f_name in ["slug", "overwrite_url"]:
                    fields.remove(f_name)
        return fields

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        if obj:
            version = Version.objects.get_for_content(obj)
            if not version.check_modify.as_bool(request.user):
                for f_name in ["slug", "overwrite_url"]:
                    form.declared_fields[f_name].widget.attrs["readonly"] = True
        return form

    def get_queryset(self, request):
        urls = ("cms_pagecontent_get_tree",)
        queryset = super().get_queryset(request)
        if request.resolver_match.url_name in urls:
            versionable = versionables.for_content(queryset.model)

            # TODO: Improve the grouping filters to use anything defined in the
            #       apps versioning config extra_grouping_fields
            grouping_filters = {}
            if "language" in versionable.extra_grouping_fields:
                grouping_filters["language"] = get_language_from_request(request)

            return queryset.filter(pk__in=versionable.distinct_groupers(**grouping_filters))
        return queryset

    # CAVEAT:
    #   - PageContent contains the template, this can differ for each language,
    #     it is assumed that templates would be the same when copying from one language to another
    # FIXME: The long term solution will require knowing:
    #           - why this view is an ajax call
    #           - where it should live going forwards (cms vs versioning)
    #           - A better way of making the feature extensible / modifiable for versioning
    def copy_language(self, request, object_id):
        target_language = request.POST.get("target_language")

        # CAVEAT: Avoiding self.get_object because it sets the page cache,
        #         We don't want a draft showing to a regular site visitor!
        #         source_page_content = self.get_object(request, object_id=object_id)
        source_page_content = PageContent._original_manager.get(pk=object_id)

        if source_page_content is None:
            raise self._get_404_exception(object_id)

        page = source_page_content.page

        if not target_language or target_language not in get_language_list(site_id=page.node.site_id):
            return HttpResponseBadRequest(force_str(_("Language must be set to a supported language!")))

        target_page_content = get_latest_admin_viewable_content(page, language=target_language)

        # First check that we are able to edit the target
        if not self.has_change_permission(request, obj=target_page_content):
            raise PermissionDenied

        for placeholder in source_page_content.get_placeholders():
            # Try and get a matching placeholder, only if it exists
            try:
                target = target_page_content.get_placeholders().get(slot=placeholder.slot)
            except ObjectDoesNotExist:
                continue

            plugins = placeholder.get_plugins_list(source_page_content.language)

            if not target.has_add_plugins_permission(request.user, plugins):
                return HttpResponseForbidden(force_str(_("You do not have permission to copy these plugins.")))
            copy_plugins_to_placeholder(plugins, target, language=target_language)
        return HttpResponse("ok")

    def change_innavigation(self, request, object_id):
        page_content = self.get_object(request, object_id=object_id)
        version = Version.objects.get_for_content(page_content)
        try:
            version.check_modify(request.user)
        except ConditionFailed as e:
            # Send error message
            return HttpResponseForbidden(force_str(e))
        return super().change_innavigation(request, object_id)

    @property
    def indicator_descriptions(self):
        """Publish indicator description to CMSPageAdmin"""
        return INDICATOR_DESCRIPTIONS

    @classmethod
    def get_indicator_menu(cls, request, page_content):
        """Get the indicator menu for PageContent object taking into account the
        currently available versions"""
        menu_template = "admin/cms/page/tree/indicator_menu.html"
        status = page_content.content_indicator()
        if not status or status == "empty":  # pragma: no cover
            return super().get_indicator_menu(request, page_content)
        versions = page_content._version  # Cache from .content_indicator()
        back = admin_reverse("cms_pagecontent_changelist") + f"?language={request.GET.get('language')}"
        menu = indicators.content_indicator_menu(request, status, versions, back=back)
        return menu_template if menu else "", menu


class VersioningCMSConfig(CMSAppConfig):
    """Implement versioning for core cms models
    """
    cms_enabled = True
    djangocms_versioning_enabled = getattr(
        settings, "VERSIONING_CMS_MODELS_ENABLED", True
    )
    versioning = [
        VersionableItem(
            content_model=PageContent,
            grouper_field_name="page",
            extra_grouping_fields=["language"],
            version_list_filter_lookups={"language": get_language_tuple},
            copy_function=copy_page_content,
            grouper_selector_option_label=label_from_instance,
            on_publish=on_page_content_publish,
            on_unpublish=on_page_content_unpublish,
            on_draft_create=on_page_content_draft_create,
            on_archive=on_page_content_archive,
            content_admin_mixin=VersioningCMSPageAdminMixin,
        )
    ]
    cms_toolbar_mixin = CMSToolbarVersioningMixin
    PageContent.add_to_class("is_editable", is_editable)
    PageContent.add_to_class("content_indicator", indicators.content_indicator)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-08-13 13:07
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("djangocms_versioning", "0006_auto_20180809_1714"),
    ]

    operations = [
        migrations.AddField(
            model_name="version",
            name="created_by",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
            ),
            preserve_default=False,
        )
    ]



from cms.app_base import CMSAppConfig

from .models import FancyPoll
from .views import render_content


class FancyPollCMSConfig(CMSAppConfig):
    cms_enabled = True
    cms_toolbar_enabled_models = [(FancyPoll, render_content)]
    djangocms_versioning_enabled = False
    versioning = []



from unittest.mock import patch

from django.core.checks import messages

from djangocms_versioning import constants
from djangocms_versioning.models import StateTracking, Version
from djangocms_versioning.test_utils import factories
from djangocms_versioning.test_utils.blogpost.cms_config import BlogpostCMSConfig
from djangocms_versioning.test_utils.polls.cms_config import PollsCMSConfig
from tests.test_admin import BaseStateTestCase


class PermissionTestCase(BaseStateTestCase):
    def setUp(self):
        self.versionable = BlogpostCMSConfig.versioning[0]
        self.poll_versionable = PollsCMSConfig.versioning[0]

    def get_user(self, username, is_staff=True):
        user = factories.UserFactory(username=username, is_staff=is_staff)
        user.set_password(username)
        user.save()
        return user

    @patch("django.contrib.messages.add_message")
    def test_publish_view_cannot_be_accessed_without_permission(
        self, mocked_messages
    ):
        post_version = factories.BlogPostVersionFactory(state=constants.DRAFT)
        url = self.get_admin_url(
            self.versionable.version_model_proxy, "publish", post_version.pk
        )

        with self.login_user_context(self.get_staff_user_with_no_permissions()):
            response = self.client.post(url)

        self.assertRedirectsToPreview(response, post_version)

        self.assertEqual(mocked_messages.call_count, 1)
        self.assertEqual(mocked_messages.call_args[0][1], messages.ERROR)
        self.assertEqual(mocked_messages.call_args[0][2], "You do not have permission to perform this action")

        # status hasn't changed
        post_version = Version.objects.get(pk=post_version.pk)
        self.assertEqual(post_version.state, constants.DRAFT)
        # no status change has been tracked
        self.assertEqual(StateTracking.objects.all().count(), 0)

    @patch("django.contrib.messages.add_message")
    def test_unpublish_view_cannot_be_accessed_without_permission(
        self, mocked_messages
    ):
        post_version = factories.BlogPostVersionFactory(state=constants.PUBLISHED)
        url = self.get_admin_url(
            self.versionable.version_model_proxy, "unpublish", post_version.pk
        )

        with self.login_user_context(self.get_staff_user_with_no_permissions()):
            response = self.client.post(url)

        self.assertRedirectsToPreview(response, post_version)

        self.assertEqual(mocked_messages.call_count, 1)
        self.assertEqual(mocked_messages.call_args[0][1], messages.ERROR)
        self.assertEqual(mocked_messages.call_args[0][2], "You do not have permission to perform this action")

        # status hasn't changed
        post_version = Version.objects.get(pk=post_version.pk)
        self.assertEqual(post_version.state, constants.PUBLISHED)
        # no status change has been tracked
        self.assertEqual(StateTracking.objects.all().count(), 0)


    @patch("django.contrib.messages.add_message")
    def test_publish_view_can_be_accessed_with_low_level_permission(
        self, mocked_messages
    ):
        # alice has no permission to publish bob's post
        post_version = factories.BlogPostVersionFactory(state=constants.DRAFT, content__text="bob's post")
        url = self.get_admin_url(
            self.versionable.version_model_proxy, "publish", post_version.pk
        )

        with self.login_user_context(self.get_user("bob")):
            self.client.post(url)

        self.assertEqual(mocked_messages.call_count, 1)
        self.assertEqual(mocked_messages.call_args[0][1], messages.INFO)
        self.assertEqual(mocked_messages.call_args[0][2], "Version published")

        # status has changed
        post_version = Version.objects.get(pk=post_version.pk)
        self.assertEqual(post_version.state, constants.PUBLISHED)
        # status change has been tracked
        self.assertEqual(StateTracking.objects.all().count(), 1)

    @patch("django.contrib.messages.add_message")
    def test_publish_view_cannot_be_accessed_wo_low_level_permission(
        self, mocked_messages
    ):
        # alice has no permission to publish bob's post
        post_version = factories.BlogPostVersionFactory(state=constants.DRAFT, content__text="bob's post")
        url = self.get_admin_url(
            self.versionable.version_model_proxy, "publish", post_version.pk
        )

        with self.login_user_context(self.get_user("alice")):
            response = self.client.post(url)

        self.assertRedirectsToPreview(response, post_version)

        self.assertEqual(mocked_messages.call_count, 1)
        self.assertEqual(mocked_messages.call_args[0][1], messages.ERROR)
        self.assertEqual(mocked_messages.call_args[0][2], "You do not have permission to perform this action")

        # status hasn't changed
        post_version = Version.objects.get(pk=post_version.pk)
        self.assertEqual(post_version.state, constants.DRAFT)
        # no status change has been tracked
        self.assertEqual(StateTracking.objects.all().count(), 0)

    @patch("django.contrib.messages.add_message")
    def test_unpublish_view_can_be_accessed_with_low_level_permission(
        self, mocked_messages
    ):
        # bob has permission to unpublish bob's post
        post_version = factories.BlogPostVersionFactory(state=constants.PUBLISHED, content__text="bob's post")
        url = self.get_admin_url(
            self.versionable.version_model_proxy, "unpublish", post_version.pk
        )

        with self.login_user_context(self.get_user("bob")):
            self.client.post(url)

        self.assertEqual(mocked_messages.call_count, 1)
        self.assertEqual(mocked_messages.call_args[0][1], messages.INFO)
        self.assertEqual(mocked_messages.call_args[0][2], "Version unpublished")

        # status has changed
        post_version = Version.objects.get(pk=post_version.pk)
        self.assertEqual(post_version.state, constants.UNPUBLISHED)
        # status change has been tracked
        self.assertEqual(StateTracking.objects.all().count(), 1)

    @patch("django.contrib.messages.add_message")
    def test_unpublish_view_cannot_be_accessed_wo_low_level_permission(
        self, mocked_messages
    ):
        # alice has no permission to unpublish bob's post
        post_version = factories.BlogPostVersionFactory(state=constants.PUBLISHED, content__text="bob's post")
        url = self.get_admin_url(
            self.versionable.version_model_proxy, "unpublish", post_version.pk
        )

        with self.login_user_context(self.get_user("alice")):
            response = self.client.post(url)

        self.assertRedirectsToPreview(response, post_version)

        self.assertEqual(mocked_messages.call_count, 1)
        self.assertEqual(mocked_messages.call_args[0][1], messages.ERROR)
        self.assertEqual(mocked_messages.call_args[0][2], "You do not have permission to perform this action")

        # status hasn't changed
        post_version = Version.objects.get(pk=post_version.pk)
        self.assertEqual(post_version.state, constants.PUBLISHED)
        # no status change has been tracked
        self.assertEqual(StateTracking.objects.all().count(), 0)

    @patch("django.contrib.messages.add_message")
    def test_archive_view_cannot_be_accessed_without_permission(
        self, mocked_messages
    ):
        post_version = factories.BlogPostVersionFactory(state=constants.DRAFT)
        url = self.get_admin_url(
            self.versionable.version_model_proxy, "archive", post_version.pk
        )
        user = self.get_staff_user_with_no_permissions()

        with self.login_user_context(user):
            self.client.post(url)

        self.assertEqual(mocked_messages.call_count, 1)
        self.assertEqual(mocked_messages.call_args[0][1], messages.ERROR)
        self.assertEqual(mocked_messages.call_args[0][2], "You do not have permission to perform this action")

        # status hasn't changed
        post_version = Version.objects.get(pk=post_version.pk)
        self.assertEqual(post_version.state, constants.DRAFT)
        # no status change has been tracked
        self.assertEqual(StateTracking.objects.all().count(), 0)

    @patch("django.contrib.messages.add_message")
    def test_archive_view_can_be_accessed_with_permission(
        self, mocked_messages
    ):
        poll_version = factories.PollVersionFactory(state=constants.DRAFT)
        url = self.get_admin_url(
            self.poll_versionable.version_model_proxy, "archive", poll_version.pk
        )
        user = self.get_staff_user_with_no_permissions()
        user.user_permissions.add(self.get_permission("change_pollcontent"))

        with self.login_user_context(user):
            self.client.post(url)

        self.assertEqual(mocked_messages.call_count, 1)
        self.assertEqual(mocked_messages.call_args[0][1], messages.INFO)
        self.assertEqual(mocked_messages.call_args[0][2], "Version archived")

        # status has changed
        poll_version_ = Version.objects.get(pk=poll_version.pk)
        self.assertEqual(poll_version_.state, constants.ARCHIVED)
        # status change has been tracked
        self.assertEqual(StateTracking.objects.all().count(), 1)

    @patch("django.contrib.messages.add_message")
    def test_revert_view_cannot_be_accessed_without_permission(
        self, mocked_messages
    ):
        post_version = factories.BlogPostVersionFactory(state=constants.ARCHIVED)
        url = self.get_admin_url(
            self.versionable.version_model_proxy, "revert", post_version.pk
        )
        user = self.get_staff_user_with_no_permissions()

        with self.login_user_context(user):
            self.client.post(url)

        self.assertEqual(mocked_messages.call_count, 1)
        self.assertEqual(mocked_messages.call_args[0][1], messages.ERROR)
        self.assertEqual(mocked_messages.call_args[0][2], "You do not have permission to perform this action")

        # status hasn't changed
        poll_version_ = Version.objects.get(pk=post_version.pk)
        self.assertEqual(poll_version_.state, constants.ARCHIVED)
        # no status change has been tracked
        self.assertEqual(StateTracking.objects.all().count(), 0)

    @patch("django.contrib.messages.add_message")
    def test_revert_view_can_be_accessed_with_low_level_permission(
        self, mocked_messages
    ):
        post_version = factories.BlogPostVersionFactory(state=constants.ARCHIVED, content__text="post <alice>")
        url = self.get_admin_url(
            self.versionable.version_model_proxy, "revert", post_version.pk
        )
        user = self.get_user("alice", is_staff=True)
        with self.login_user_context(user):
            self.client.post(url)

        # new draft has been created
        post_version_ = Version.objects.filter(
            content_type=post_version.content_type,
            object_id__gt=post_version.object_id,
            pk__gt=post_version.pk
        ).first()
        self.assertIsNotNone(post_version_)
        self.assertEqual(post_version_.state, constants.DRAFT)
        self.assertTrue(post_version_.content.has_change_permission(user))  # Content was copied



from cms.cms_toolbars import LANGUAGE_MENU_IDENTIFIER, PlaceholderToolbar
from cms.test_utils.testcases import CMSTestCase
from cms.toolbar.utils import get_object_edit_url, get_object_preview_url
from cms.utils.urlutils import admin_reverse
from django.contrib.auth.models import Permission
from django.utils.text import slugify

from djangocms_versioning.cms_config import VersioningCMSConfig
from djangocms_versioning.constants import ARCHIVED, DRAFT, PUBLISHED
from djangocms_versioning.helpers import version_list_url
from djangocms_versioning.test_utils.factories import (
    BlogPostVersionFactory,
    FancyPollFactory,
    PageContentWithVersionFactory,
    PageUrlFactory,
    PageVersionFactory,
    PollVersionFactory,
    UserFactory,
)
from djangocms_versioning.test_utils.polls.cms_config import PollsCMSConfig
from djangocms_versioning.test_utils.test_helpers import (
    find_toolbar_buttons,
    get_toolbar,
    toolbar_button_exists,
)


class VersioningToolbarTestCase(CMSTestCase):
    def _get_publish_url(self, version, versionable=PollsCMSConfig.versioning[0]):
        """Helper method to return the expected publish url
        """
        admin_url = self.get_admin_url(
            versionable.version_model_proxy, "publish", version.pk
        )
        return admin_url

    def _get_edit_url(self, version, versionable=PollsCMSConfig.versioning[0]):
        """Helper method to return the expected edit redirect url
        """
        admin_url = self.get_admin_url(
            versionable.version_model_proxy, "edit_redirect", version.pk
        )
        return admin_url

    def _get_revert_url(self, version, versionable=PollsCMSConfig.versioning[0]):
        """Helper method to return the expected publish url
        """
        admin_url = self.get_admin_url(
            versionable.version_model_proxy, "revert", version.pk
        )
        return admin_url

    def test_publish_in_toolbar_in_edit_mode(self):
        """Test for Edit button in edit mode"""
        version = PollVersionFactory()
        toolbar = get_toolbar(version.content, edit_mode=True)

        toolbar.post_template_populate()
        revert_button = find_toolbar_buttons("Revert", toolbar.toolbar)
        self.assertListEqual(revert_button, [])  # No revert button

        publish_button = find_toolbar_buttons("Publish", toolbar.toolbar)[0]
        self.assertEqual(publish_button.name, "Publish")
        self.assertEqual(publish_button.url, self._get_publish_url(version))
        self.assertFalse(publish_button.disabled)
        self.assertListEqual(
            publish_button.extra_classes,
            ["cms-btn-action", "js-action", "cms-form-post-method", "cms-versioning-js-publish-btn"],
        )

    def test_revert_in_toolbar_in_preview_mode(self):
        """Test for Revert button outside mode"""

        version = PollVersionFactory()
        version.archive(self.get_superuser())
        toolbar = get_toolbar(version.content, edit_mode=False, user=self.get_superuser())

        toolbar.post_template_populate()
        publish_button = find_toolbar_buttons("Publish", toolbar.toolbar)
        self.assertListEqual(publish_button, [])  # No publish button

        revert_button = find_toolbar_buttons("Revert", toolbar.toolbar)[0]
        self.assertEqual(revert_button.name, "Revert")
        self.assertEqual(revert_button.url, self._get_revert_url(version))
        self.assertFalse(revert_button.disabled)
        self.assertListEqual(
            revert_button.extra_classes,
            ["cms-btn-action", ],
        )

    def test_publish_not_in_toolbar_in_preview_mode(self):
        version = PollVersionFactory()
        toolbar = get_toolbar(version.content, preview_mode=True)

        toolbar.post_template_populate()

        self.assertFalse(toolbar_button_exists("Publish", toolbar.toolbar))

    def test_publish_not_in_toolbar_in_structure_mode(self):
        version = PollVersionFactory()
        toolbar = get_toolbar(version.content, structure_mode=True)

        toolbar.post_template_populate()

        self.assertFalse(toolbar_button_exists("Publish", toolbar.toolbar))

    def test_dont_add_publish_for_models_not_registered_with_versioning(self):
        # User objects are not registered with versioning, so attempting
        # to populate a user toolbar should not attempt to add a publish
        # button
        toolbar = get_toolbar(UserFactory(), edit_mode=True)

        toolbar.post_template_populate()

        self.assertFalse(toolbar_button_exists("Publish", toolbar.toolbar))

    def test_url_for_publish_uses_version_id_not_content_id(self):
        """Regression test for a bug. Make sure than when we generate
        the publish url, we use the id of the version record, not the
        id of the content record.
        """
        # All versions are stored in the version table so increase the
        # id of version id sequence by creating a blogpost version
        BlogPostVersionFactory()
        # Now create a poll version - the poll content and version id
        # will be different.
        version = PollVersionFactory()
        toolbar = get_toolbar(
            version.content, user=self.get_superuser(), edit_mode=True
        )
        toolbar.post_template_populate()
        publish_button = find_toolbar_buttons("Publish", toolbar.toolbar)[0]

        self.assertEqual(publish_button.url, self._get_publish_url(version))

    def test_edit_in_toolbar_in_preview_mode(self):
        version = PageVersionFactory(content__template="")
        toolbar = get_toolbar(
            version.content, user=self.get_superuser(), preview_mode=True
        )

        toolbar.post_template_populate()
        edit_button = find_toolbar_buttons("Edit", toolbar.toolbar)[0]

        self.assertEqual(edit_button.name, "Edit")
        self.assertEqual(
            edit_button.url,
            self._get_edit_url(version, VersioningCMSConfig.versioning[0]),
        )
        self.assertFalse(edit_button.disabled)
        self.assertListEqual(
            edit_button.extra_classes,
            ["cms-btn-action", "js-action", "cms-form-post-method", "cms-versioning-js-edit-btn"]
        )

    def test_edit_not_in_toolbar_in_edit_mode(self):
        version = PollVersionFactory()
        toolbar = get_toolbar(
            version.content, user=self.get_superuser(), edit_mode=True
        )

        toolbar.post_template_populate()

        self.assertFalse(toolbar_button_exists("Edit", toolbar.toolbar))

    def test_edit_not_in_toolbar_in_structure_mode(self):
        version = PollVersionFactory()
        toolbar = get_toolbar(
            version.content, user=self.get_superuser(), structure_mode=True
        )

        toolbar.post_template_populate()

        self.assertFalse(toolbar_button_exists("Edit", toolbar.toolbar))

    def test_dont_add_edit_for_models_not_registered_with_versioning(self):
        # User objects are not registered with versioning, so attempting
        # to populate a user toolbar should not attempt to add a edit
        # button
        toolbar = get_toolbar(
            UserFactory(), user=self.get_superuser(), preview_mode=True
        )

        toolbar.post_template_populate()

        self.assertFalse(toolbar_button_exists("Edit", toolbar.toolbar))

    def test_url_for_edit_uses_version_id_not_content_id(self):
        """Regression test for a bug. Make sure than when we generate
        the edit url, we use the id of the version record, not the
        id of the content record.
        """
        # All versions are stored in the version table so increase the
        # id of version id sequence by creating a blogpost version
        BlogPostVersionFactory()
        # Now create a page version - the page content and version id
        # will be different.
        version = PageVersionFactory(content__template="")
        toolbar = get_toolbar(
            version.content, user=self.get_superuser(), preview_mode=True
        )
        edit_url = self._get_edit_url(version, VersioningCMSConfig.versioning[0])

        toolbar.post_template_populate()

        edit_button = find_toolbar_buttons("Edit", toolbar.toolbar)[0]
        self.assertEqual(edit_button.url, edit_url)

    def test_default_cms_edit_button_is_replaced_by_versioning_edit_button(self):
        """
        The versioning edit button is available on the toolbar
        when versioning is installed and the model is versionable.
        """
        page = PageVersionFactory(content__template="", content__language="en")
        url = get_object_preview_url(page.content)

        edit_url = self._get_edit_url(
            page, VersioningCMSConfig.versioning[0]
        )

        with self.login_user_context(self.get_superuser()):
            response = self.client.post(url)

        found_button_list = find_toolbar_buttons("Edit", response.wsgi_request.toolbar)
        # Only one edit button exists
        self.assertEqual(len(found_button_list), 1)
        # The only edit button that exists is the versioning button
        self.assertEqual(found_button_list[0].url, edit_url)

    def test_default_cms_edit_button_is_used_for_non_versioned_model(self):
        """
        The default cms edit button is present for a default model
        """
        unversionedpoll = FancyPollFactory()
        url = get_object_preview_url(unversionedpoll, language="en")
        edit_url = get_object_edit_url(unversionedpoll)

        with self.login_user_context(self.get_superuser()):
            response = self.client.post(url)

        found_button_list = find_toolbar_buttons("Edit", response.wsgi_request.toolbar)

        # Only one edit button exists
        self.assertEqual(len(found_button_list), 1)
        # The only edit button that exists is the standard cms button
        self.assertEqual(found_button_list[0].url, edit_url)

    def test_default_edit_button_from_cms_exists(self):
        """
        The default toolbar Edit button exists.
        """
        pagecontent = PageVersionFactory(content__template="")
        edit_url = self._get_edit_url(
            pagecontent.content, VersioningCMSConfig.versioning[0]
        )

        toolbar = get_toolbar(
            pagecontent.content,
            user=self.get_superuser(),
            toolbar_class=PlaceholderToolbar,
            preview_mode=True,
        )
        toolbar.post_template_populate()
        found_button_list = find_toolbar_buttons("Edit", toolbar.toolbar)

        # The only edit button that exists is the default cms button and not the versioning edit button
        self.assertEqual(len(found_button_list), 1)
        self.assertNotEqual(found_button_list[0].url, edit_url)

    def test_version_menu_for_non_version_content(self):
        # User objects are not registered with versioning, so attempting
        # to populate toolbar shouldn't contain a version menu
        toolbar = get_toolbar(UserFactory(), user=self.get_superuser(), edit_mode=True)
        toolbar.post_template_populate()
        version_menu = toolbar.toolbar.get_menu("version")
        self.assertIsNone(version_menu)

    def test_version_menu_for_version_content(self):
        # Versioned item should have versioning menu
        user = UserFactory(is_staff=True)
        user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="djangocms_versioning",
                codename="change_pollcontentversion",
            )
        )
        version = PollVersionFactory()
        toolbar = get_toolbar(version.content, user=user, preview_mode=True)
        toolbar.post_template_populate()
        version_menu = toolbar.toolbar.get_menu("version")
        self.assertEqual(version_menu.get_items()[0].name, "Manage Versions...")

    def test_version_menu_for_version_content_no_permission(self):
        """Manage versions entry shouldn't appear if user doesn't have
        access to that endpoint"""
        user = UserFactory(is_staff=True)
        version = PollVersionFactory()
        toolbar = get_toolbar(version.content, user=user, preview_mode=True)
        toolbar.post_template_populate()
        version_menu = toolbar.toolbar.get_menu("version")
        self.assertFalse(version_menu.get_items())

    def test_version_menu_for_none_version(self):
        # Version menu shouldnt be generated if version is None
        version = None
        toolbar = get_toolbar(version, user=self.get_superuser(), preview_mode=True)
        toolbar.post_template_populate()
        version_menu = toolbar.toolbar.get_menu("version")
        self.assertIsNone(version_menu)

    def test_version_menu_and_url_for_version_content(self):
        # Versioned item should have versioning menu and url should be version list url
        version = PollVersionFactory()
        toolbar = get_toolbar(
            version.content, user=self.get_superuser(), preview_mode=True
        )
        toolbar.post_template_populate()
        version_menu = toolbar.toolbar.get_menu("version")
        self.assertIsNotNone(version_menu)
        self.assertEqual(
            version_menu.get_items()[0].url, version_list_url(version.content)
        )

    def test_version_menu_label(self):
        # Versioned item should have correct version menu label
        from djangocms_versioning.constants import VERSION_STATES
        version = PollVersionFactory()
        toolbar = get_toolbar(
            version.content, user=self.get_superuser(), preview_mode=True
        )
        toolbar.post_template_populate()
        version_menu = toolbar.toolbar.get_menu("version")

        expected_label = f"Version #{version.number} ({dict(VERSION_STATES)[version.state]})"

        self.assertEqual(expected_label, version_menu.name)

    def test_view_published_in_toolbar_in_edit_mode_for_published_page(self):
        """
        The 'View Published' control is only relevant for pages that
        are published
        """
        published_version = PageVersionFactory(content__language="en", state=PUBLISHED)
        toolbar = get_toolbar(published_version.content, edit_mode=True)

        toolbar.post_template_populate()

        self.assertTrue(toolbar_button_exists("View Published", toolbar.toolbar))

    def test_view_published_in_toolbar_in_preview_mode_for_published_page(self):
        """
        The 'View Published' control is only relevant for pages that
        are published
        """
        published_version = PageVersionFactory(content__language="en", state=PUBLISHED)
        toolbar = get_toolbar(published_version.content, preview_mode=True)

        toolbar.post_template_populate()

        self.assertTrue(toolbar_button_exists("View Published", toolbar.toolbar))

    def test_view_published_not_in_toolbar_in_edit_mode_for_draft_page(self):
        """
        The 'View Published' control is only relevant for pages that
        are published
        """
        draft_version = PageVersionFactory(content__language="en")
        toolbar = get_toolbar(draft_version.content, edit_mode=True)

        toolbar.post_template_populate()

        self.assertFalse(toolbar_button_exists("View Published", toolbar.toolbar))

    def test_view_published_not_in_toolbar_in_preview_mode_for_draft_page(self):
        """
        The 'View Published' control is only relevant for pages that
        are published
        """
        draft_version = PageVersionFactory(content__language="en")
        toolbar = get_toolbar(draft_version.content, preview_mode=True)

        toolbar.post_template_populate()

        self.assertFalse(toolbar_button_exists("View Published", toolbar.toolbar))

    def test_view_published_not_in_toolbar_in_edit_mode_for_poll(self):
        """
        The 'View Published' toolbar control is only relevant for pages that have
        the concept of a live url / web viewable url with the toolbar
        """
        version = PollVersionFactory(state=PUBLISHED)
        toolbar = get_toolbar(version.content, edit_mode=True)

        toolbar.post_template_populate()

        self.assertFalse(toolbar_button_exists("View Published", toolbar.toolbar))

    def test_view_published_not_in_toolbar_in_preview_mode_for_poll(self):
        """
        The 'View Published' toolbar control is only relevant for pages that have
        the concept of a live url / web viewable url with the toolbar
        """
        version = PollVersionFactory(state=PUBLISHED)
        toolbar = get_toolbar(version.content, preview_mode=True)

        toolbar.post_template_populate()

        self.assertFalse(toolbar_button_exists("View Published", toolbar.toolbar))

    def test_view_published_in_toolbar_in_edit_mode_button_url(self):
        """
        The 'View Published' toolbar control url should be a valid
        live url when in the edit mode
        """
        published_version = PageVersionFactory(content__language="en")
        language = published_version.content.language
        PageUrlFactory(
            page=published_version.content.page,
            language=language,
            path=slugify("test_page"),
            slug=slugify("test_page"),
        )
        published_version.publish(user=self.get_superuser())
        draft_version = published_version.copy(self.get_superuser())
        edit_endpoint = get_object_edit_url(draft_version.content, language="en")
        expected_url = published_version.content.page.get_absolute_url(language=language)

        with self.login_user_context(self.get_superuser()):
            response = self.client.get(edit_endpoint)

        found_button_list = find_toolbar_buttons("View Published", response.wsgi_request.toolbar)

        # check only one View Published button exists
        self.assertEqual(len(found_button_list), 1)
        self.assertEqual(found_button_list[0].url, expected_url)

    def test_view_published_in_toolbar_in_preview_mode_button_url(self):
        """
        The 'View Published' toolbar control url should be a valid
        live url when in the preview mode
        """
        published_version = PageVersionFactory(content__language="en")
        language = published_version.content.language
        PageUrlFactory(
            page=published_version.content.page,
            language=language,
            path=slugify("test_page"),
            slug=slugify("test_page"),
        )
        published_version.publish(user=self.get_superuser())
        draft_version = published_version.copy(self.get_superuser())
        preview_endpoint = get_object_preview_url(draft_version.content, language="en")
        expected_url = published_version.content.page.get_absolute_url(language=language)

        with self.login_user_context(self.get_superuser()):
            response = self.client.get(preview_endpoint)

        found_button_list = find_toolbar_buttons("View Published", response.wsgi_request.toolbar)

        # check only one View Published button exists
        self.assertEqual(len(found_button_list), 1)
        self.assertEqual(found_button_list[0].url, expected_url)


class VersioningPageToolbarTestCase(CMSTestCase):

    def _get_toolbar_item_by_name(self, menu, name):
        for item in menu.items:
            if hasattr(item, "name") and item.name == name:
                return item
        return None

    def test_change_language_menu_page_toolbar(self):
        """Check that patched PageToolbar.change_language_menu only provides
        Add Translation links.
        """
        version = PageVersionFactory(content__language="en")
        PageContentWithVersionFactory(page=version.content.page, language="de")
        PageContentWithVersionFactory(page=version.content.page, language="it")
        page = version.content.page
        page.update_languages(["en", "de", "it"])

        request = self.get_page_request(
            page=page,
            path=get_object_edit_url(version.content),
            user=self.get_superuser(),
        )
        request.toolbar.set_object(version.content)
        request.toolbar.populate()
        request.toolbar.post_template_populate()

        language_menu = request.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER)
        # 3 out of 4 populated languages, Break, Add Translation menu, Copy all plugins
        self.assertEqual(language_menu.get_item_count(), 6)

        language_menu_dict = {
            menu.name: list(menu.items)
            for key, menu in language_menu.menus.items()
        }
        self.assertIn("Add Translation", language_menu_dict.keys())
        self.assertIn("Copy all plugins", language_menu_dict.keys())
        self.assertNotIn("Delete Translation", language_menu_dict.keys())

        self.assertEqual(
            {lang.name for lang in language_menu_dict["Add Translation"]},
            {"Française..."},
        )

        self.assertEqual(
            {lang.name for lang in language_menu_dict["Copy all plugins"]},
            {"from Italiano", "from Deutsche"},
        )

        for item in language_menu_dict["Add Translation"]:
            self.assertIn(admin_reverse("cms_pagecontent_add"), item.url)
            self.assertIn(f"cms_page={page.pk}", item.url)
            lang_code = "fr" if "Française" in item.name else "it"
            self.assertIn(f"language={lang_code}", item.url)

    def test_change_language_menu_page_toolbar_language_selector_version_link(self):
        """
        Ensure that the correct version is navigated to in the language selector.

        A real world scenario / issue seen:
            - Version 3: Draft
            - Version 2: Published
            - Version 1: Archived

        Version 1 was returned in the toolbar language selector which is incorrect,
        the latest version 4 should be returned.
        """
        superuser = self.get_superuser()
        en_pagecontent_1 = PageContentWithVersionFactory(language="en")
        page = en_pagecontent_1.page
        de_pagecontent_1 = PageContentWithVersionFactory(page=page, language="de")
        # Create remaining 3 versions for it
        it_pagecontent_1 = PageContentWithVersionFactory(page=page, language="it", version__state=ARCHIVED)
        it_pagecontent_1_version = it_pagecontent_1.versions.first()
        # Make version 1 archived by publishing the new version 2
        it_pagecontent_2 = PageContentWithVersionFactory(page=page, language="it", version__state=PUBLISHED)
        it_pagecontent_2_version = it_pagecontent_2.versions.first()
        # Create a new draft, which is what we expect to use
        it_pagecontent_3 = PageContentWithVersionFactory(page=page, language="it", version__state=DRAFT)
        it_pagecontent_3_version = it_pagecontent_3.versions.first()

        # Sanity check that all versions are int he state that we expect: Archived, Published, Draft
        self.assertEqual(it_pagecontent_1_version.state, ARCHIVED)
        self.assertEqual(it_pagecontent_2_version.state, PUBLISHED)
        self.assertEqual(it_pagecontent_3_version.state, DRAFT)

        page.update_languages(["en", "de", "it"])

        request = self.get_page_request(
            page=page,
            path=get_object_edit_url(en_pagecontent_1),
            user=superuser,
        )
        request.toolbar.set_object(en_pagecontent_1)
        request.toolbar.populate()
        request.toolbar.post_template_populate()

        language_menu = request.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER)
        language_menu_item_names = [item.name for item in language_menu.items if hasattr(item, "name")]

        self.assertIn("English", language_menu_item_names)
        self.assertIn("Deutsche", language_menu_item_names)
        self.assertIn("Italiano", language_menu_item_names)

        en_item = self._get_toolbar_item_by_name(language_menu, "English")
        en_preview_url = get_object_preview_url(en_pagecontent_1, "en")
        de_item = self._get_toolbar_item_by_name(language_menu, "Deutsche")
        de_preview_url = get_object_preview_url(de_pagecontent_1, "de")
        it_item = self._get_toolbar_item_by_name(language_menu, "Italiano")
        it_preview_url = get_object_preview_url(it_pagecontent_3, "it")

        # Ensure that each menu item points to the correct url
        self.assertEqual(en_item.url, en_preview_url)
        self.assertEqual(de_item.url, de_preview_url)
        self.assertEqual(it_item.url, it_preview_url)

    def test_page_toolbar_wo_language_menu(self):
        from django.utils.translation import gettext as _

        pagecontent = PageContentWithVersionFactory(language="en")
        page = pagecontent.page
        # Get request
        request = self.get_page_request(
            page=page,
            path=get_object_edit_url(pagecontent),
            user=self.get_superuser(),
        )
        # Remove language menu from request's toolbar
        del request.toolbar.menus[LANGUAGE_MENU_IDENTIFIER]

        # find VersioningPageToolbar
        for cls, toolbar in request.toolbar.toolbars.items():
            if cls == "djangocms_versioning.cms_toolbars.VersioningPageToolbar":
                # and call override_language_menu
                toolbar.override_language_menu()
                break

        language_menu = request.toolbar.get_menu(LANGUAGE_MENU_IDENTIFIER, _("Language"))
        self.assertIsNone(language_menu)



from cms.models.fields import PlaceholderRelationField
from cms.test_utils.testcases import CMSTestCase

from djangocms_versioning.constants import (
    ARCHIVED,
    DRAFT,
    PUBLISHED,
    UNPUBLISHED,
)
from djangocms_versioning.helpers import is_content_editable
from djangocms_versioning.test_utils.factories import (
    FancyPollFactory,
    PageVersionFactory,
    PlaceholderFactory,
)


class CheckDraftEditableTestCase(CMSTestCase):
    def test_check_archived_state(self):
        user = self.get_superuser()
        version = PageVersionFactory(state=ARCHIVED)
        placeholder = PlaceholderFactory(source=version.content)
        self.assertFalse(is_content_editable(placeholder, user))

    def test_check_draft_state(self):
        user = self.get_superuser()
        version = PageVersionFactory(state=DRAFT)
        placeholder = PlaceholderFactory(source=version.content)
        self.assertTrue(is_content_editable(placeholder, user))

    def test_check_unpublished_state(self):
        user = self.get_superuser()
        version = PageVersionFactory(state=UNPUBLISHED)
        placeholder = PlaceholderFactory(source=version.content)
        self.assertFalse(is_content_editable(placeholder, user))

    def test_check_published_state(self):
        user = self.get_superuser()
        version = PageVersionFactory(state=PUBLISHED)
        placeholder = PlaceholderFactory(source=version.content)
        self.assertFalse(is_content_editable(placeholder, user))

    def test_check_unversioned_model(self):
        user = self.get_superuser()
        placeholder = PlaceholderFactory(source=FancyPollFactory())
        self.assertTrue(is_content_editable(placeholder, user))


class CheckInjectTestCase(CMSTestCase):
    def test_draft_state_check_is_injected_into_default_checks(self):
        self.assertIn(is_content_editable, PlaceholderRelationField.default_checks)
